/*
  KeePT - Plugin for KeePass 1.21
  Copyright (C) 2010 Nikolaus Hammler <niki@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <tchar.h>
#include <assert.h>
#include "KeePTPlugin.h"
#include "WinPTHooks.h"
#include "WinPTExec.h"
#include "KeePTHotKeys.h"

IKpAPI3 *g_pAPI = NULL;
HINSTANCE g_hDLL = NULL;
static WNDPROC g_pOldWndProc = NULL;
UINT WM_HKEYCNTRL = 0;

#ifdef _MSC_VER
#pragma data_seg(".SHARED")
static BOOL g_shbOneInstance = FALSE;
#pragma data_seg()
#pragma comment(linker, "/section:.SHARED,RWS")
#else
#define SHARED __attribute__((section(".SHARED"), shared)) 
static BOOL g_shbOneInstance SHARED = FALSE;
#endif

#ifdef _MSC_VER
KP_IMPL_STDREFIMPL(KeePTPlugin)
#else
STDMETHODIMP_(ULONG) KeePTPlugin::AddRef()
{
	return ++__m_uInterfaceRefCount;
}

STDMETHODIMP_(ULONG) KeePTPlugin::Release()
{
	ULONG __kp_ref = __m_uInterfaceRefCount;
	if(__m_uInterfaceRefCount != 0)
	{
		__kp_ref = --__m_uInterfaceRefCount;
		if(__m_uInterfaceRefCount == 0) delete this;
	}
	else
	{
		assert(false);
	}
	return __kp_ref;
}
#endif

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	UNREFERENCED_PARAMETER(lpvReserved);
	g_hDLL = hinstDLL;

	switch(fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			return InjectPT(hinstDLL);
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
	}
	
	return TRUE;
}

KP_EXPORT HRESULT KP_API KP_I_INITIALIZELIB_DECL(IKpUnknown* pAPI)
{
	UNREFERENCED_PARAMETER(pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_RELEASELIB_DECL(IKpUnknown* pAPI)
{
	UNREFERENCED_PARAMETER(pAPI);
	SAFE_RELEASE(g_pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_CREATEINSTANCE_DECL(REFIID riid, void** ppvObject, IKpUnknown* pAPI)
{
	*ppvObject = NULL;

	if(pAPI == NULL) return E_UNEXPECTED;

	if((riid == IID_IKpUnknown) || (riid == IID_IKpPlugin))
	{
		if(pAPI->QueryInterface(IID_IKpAPI3, (void**)&g_pAPI) != S_OK)
			return E_FAIL;

		if(g_shbOneInstance)
		{
			MessageBox(NULL, _T("This plugin is only supported in one KeePass instance. KeePT.dll will not load."), MSGTITLE, MB_OK | MB_ICONWARNING);
			return E_FAIL;
		}
		// TODO: We could use a kind of Mutex here
		// However, the chance that two plugins are loaded at
		// the same time is very low and the side effects very small
		g_shbOneInstance = TRUE;

		IKpPlugin* pImpl = new KeePTPlugin();
		*ppvObject = ((riid == IID_IKpPlugin) ? pImpl : (IKpUnknown*)pImpl);
		return S_OK;
	}

	return E_NOINTERFACE;
}

KeePTPlugin::KeePTPlugin()
{
	KP_IMPL_CONSTRUCT;

	ZeroMemory(&m_menuItems[0], sizeof(KP_MENU_ITEM) * MENUITEMCOUNT);

	m_menuItems[0].lpCommandString = (LPTSTR)_T("KeePT...");
}

KeePTPlugin::~KeePTPlugin()
{
	g_shbOneInstance = FALSE;
}

#define KP_SUPPORT_INTERFACE(__kp_iid,__kp_int) { assert(__m_uInterfaceRefCount > 0); \
	if(riid == __kp_iid) { __kp_int##* __kp_p = this; *ppvObject = __kp_p; \
	this->AddRef(); return S_OK; } }

STDMETHODIMP KeePTPlugin::QueryInterface(REFIID riid, void** ppvObject)
{
#ifdef _MSC_VER
	KP_SUPPORT_INTERFACE(IID_IKpUnknown, IKpUnknown);
	KP_SUPPORT_INTERFACE(IID_IKpPlugin, IKpPlugin);
#else
	{
		assert(__m_uInterfaceRefCount > 0);
		if(riid == IID_IKpUnknown)
		{
			IKpUnknown* __kp_p = this;
			*ppvObject = __kp_p;
			this->AddRef(); return S_OK;
		}
	}
	{
		assert(__m_uInterfaceRefCount > 0);
		if(riid == IID_IKpPlugin)
		{
			IKpPlugin* __kp_p = this;
			*ppvObject = __kp_p;
			this->AddRef(); return S_OK;
		}
	}
#endif

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

BOOL CALLBACK SettingsDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	IKpConfig *pConf = NULL;
	IKpUtilities *pUtils = NULL;

	switch(uMsg)
	{
		case WM_INITDIALOG:
			if(g_pAPI->CreateInstance(CLSID_KpConfig_ReadOnly,
				IID_IKpConfig, (void**)&pConf) != S_OK)
				return FALSE;

			assert( pConf );

			char szPath[SI_REGSIZE];
			pConf->Get(CONF_WPTPATH, szPath);
			SetDlgItemText(hwndDlg, IDC_WINPT, szPath);

			CheckDlgButton(hwndDlg, IDC_CHECK_CLOSEPT,
				pConf->GetBool(CONF_WPTCLOSE, CONF_WPTCLOSE_DEF) ? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hwndDlg, IDC_CHECK_HKEYS,
				pConf->GetBool(CONF_WPTHKEY, CONF_WPTHKEY_DEF) ? BST_CHECKED : BST_UNCHECKED);

			CheckRadioButton(hwndDlg, IDC_WINPT_STANDARD, IDC_WINPT_PORTABLE,
				pConf->GetBool(CONF_WPTUSEPORTABLE, CONF_WPTUSEPORTABLE_DEF) ? IDC_WINPT_PORTABLE : IDC_WINPT_STANDARD);

			pConf->Release();
			return TRUE;
		case WM_COMMAND:
			switch(wParam)
			{
				case IDCANCEL:
					EndDialog(hwndDlg, 0);
					return TRUE;
				case IDOK:
					if(g_pAPI->CreateInstance(CLSID_KpConfig,
						IID_IKpConfig, (void**)&pConf) != S_OK)
						return FALSE;

					assert( pConf );

					char szPath[SI_REGSIZE];
					GetDlgItemText(hwndDlg, IDC_WINPT, szPath, SI_REGSIZE-1);
					pConf->Set(CONF_WPTPATH, szPath);

					pConf->SetBool(CONF_WPTCLOSE, IsDlgButtonChecked(hwndDlg, IDC_CHECK_CLOSEPT) == BST_CHECKED ? TRUE : FALSE);
					pConf->SetBool(CONF_WPTHKEY, IsDlgButtonChecked(hwndDlg, IDC_CHECK_HKEYS) == BST_CHECKED ? TRUE : FALSE);

					pConf->SetBool(CONF_WPTUSEPORTABLE, IsDlgButtonChecked(hwndDlg, IDC_WINPT_PORTABLE) == BST_CHECKED ? TRUE : FALSE);

					pConf->Release();
					EndDialog(hwndDlg, 0);
					return TRUE;
				case IDC_BROWSE:
					if(g_pAPI->QueryInstance(SCLSID_KpUtilities, IID_IKpUtilities, (void**)&pUtils) != S_OK)
						return FALSE;
					assert( pUtils );

					char szPath1[SI_REGSIZE];
					if(pUtils->ShowFileDialog(TRUE, "exe", szPath1, SI_REGSIZE-1) == S_OK)
						SetDlgItemText(hwndDlg, IDC_WINPT, szPath1);
					pUtils->Release();
					return TRUE;
			}
	}
	return FALSE;
}

LRESULT CALLBACK KeePassWinProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	assert( g_pOldWndProc != 0 );

	if(WM_HKEYCNTRL && msg == WM_HKEYCNTRL)
	{
		if(wParam == WM_HKEYCNTRL_UPDATE)
			UpdateHotKeys(hwnd);
		if(wParam == WM_HKEYCNTRL_UNREGISTER)
			UnregisterHotKeys(hwnd);
		return 0;
	}

	switch(msg)
	{
		case WM_COPYDATA:
			{
				LRESULT ret = FALSE;
				COPYDATASTRUCT *cds = (COPYDATASTRUCT*)lParam;
				if((ret = OnCopyData(hwnd, (HWND)wParam, cds)) != FALSE)
					return ret;
			}
			break;
		case WM_HOTKEY:
			if(OnWmHotKey((int)wParam, (UINT) LOWORD(lParam), (UINT) HIWORD(lParam)))
				return 0;
			break;
	}

	return g_pOldWndProc(hwnd, msg, wParam, lParam);
}

void KeePTSettings()
{
	assert( g_pAPI );
	DialogBox(g_hDLL, MAKEINTRESOURCE(IDD_SETTINGS), g_pAPI->GetMainWindowHandle(), SettingsDlgProc);
	UpdateHotKeys(g_pAPI->GetMainWindowHandle());
}

STDMETHODIMP_(BOOL) KeePTPlugin::OnMessage(DWORD dwCode, LPARAM lParamW, LPARAM lParamL)
{
	switch(dwCode)
	{
		case KPM_DELAYED_INIT:
			WM_HKEYCNTRL = RegisterWindowMessage(WM_HKEYCNTRL_GUID);
#ifdef _MSC_VER
#pragma warning(disable:4244 4312)
#endif
			g_pOldWndProc = (WNDPROC)SetWindowLongPtr(g_pAPI->GetMainWindowHandle(), GWLP_WNDPROC, (LONG_PTR)KeePassWinProc);
#ifdef _MSC_VER
#pragma warning(default:4244 4312)
#endif
			UpdateHotKeys(g_pAPI->GetMainWindowHandle());
			break;
		case KPM_CLEANUP:
			UnregisterHotKeys(g_pAPI->GetMainWindowHandle());
#ifdef _MSC_VER
#pragma warning(disable:4244)
#endif
			SetWindowLongPtr(g_pAPI->GetMainWindowHandle(), GWLP_WNDPROC, (LONG_PTR)g_pOldWndProc);
			g_pOldWndProc = NULL;
#ifdef _MSC_VER
#pragma warning(default:4244)
#endif
			break;
		case KPM_DIRECT_CONFIG:
			KeePTSettings();
			return TRUE;
		//case KPM_TRAY_NOTIFY:
		//	return OnKpmTrayNotify(lParamW, lParamL);
		case KPM_TRAY_CUSTOMIZE_PRE:
			{
				HMENU hMenu = (HMENU)lParamW;
				if(!hMenu)
					break;

				MENUITEMINFO item;
				ZeroMemory(&item, sizeof(MENUITEMINFO));
				item.cbSize = sizeof(MENUITEMINFO);
				
				item.fMask = MIIM_ID;
				item.wID = m_menuItems[0].dwCommandID;

				item.fMask |= MIIM_STRING;
				item.dwTypeData = (LPSTR)_T("KeePT...\t->");

				InsertMenuItem(hMenu, 1, TRUE, &item);


				HBITMAP hMenuIcon = LoadBitmap(g_hDLL, MAKEINTRESOURCE(IDB_MENUICON));
				if(hMenuIcon)
				{
					SetMenuItemBitmaps(hMenu, 1, MF_BYPOSITION, hMenuIcon, NULL);
				}
			}
			break;
		case KPM_DIRECT_EXEC:
			if((DWORD)lParamW == m_menuItems[0].dwCommandID)
			{
				if(!ShowWinPTTrayMenu())
				{
					assert( g_pAPI );
					if(MessageBox(g_pAPI->GetMainWindowHandle(), _T("Could not start WinPT or WinPT is unsupported. Configure the plugin?"), MSGTITLE, MB_YESNO | MB_ICONERROR) == IDYES)
					{
						KeePTSettings();
						return TRUE;
					}
					return FALSE;
				}
			}
			break;
		case KPM_PLUGIN_INFO:
			assert( g_pAPI );
			MessageBox(g_pAPI->GetMainWindowHandle(), _T("(c) 2011 Nikolaus Hammler <niki@hammler.net>\n\nhttp://niki.hammler.net"), MSGTITLE, MB_OK | MB_ICONINFORMATION);
			return TRUE;
		default:
			break;
	}

	return TRUE;
}

STDMETHODIMP_(LPCTSTR) KeePTPlugin::GetProperty(LPCTSTR lpName)
{
	// unload the plugin late so that we can subclass the main window
	// (otherwise, the plugin would be unloaded while our KeePassWinProc
	// is on the call stack and KeePass would crash
	static char ret[] = "True";
	if(strcmp(lpName, KPPS_UNLOAD_LATE) == 0)
		return ret;
	return NULL;
}

STDMETHODIMP KeePTPlugin::SetProperty(LPCTSTR lpName, LPCTSTR lpValue)
{
	UNREFERENCED_PARAMETER(lpName);
	UNREFERENCED_PARAMETER(lpValue);
	return E_FAIL;
}

STDMETHODIMP_(DWORD) KeePTPlugin::GetMenuItemCount()
{
	return MENUITEMCOUNT;
}

STDMETHODIMP_(KP_MENU_ITEM*) KeePTPlugin::GetMenuItems()
{
	return m_menuItems;
}

